"""
This type stub file was generated by pyright.
"""

class GitUrlParsed:
    def __init__(self, parsed_info): ...
    @property
    def valid(self) -> bool: ...
    @property
    def host(self) -> str: ...
    @property
    def resource(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def user(self) -> str: ...
    @property
    def pathname(self) -> str: ...
    @property
    def platform(self) -> str: ...

def parse(url: str, check_domain: bool = ...) -> GitUrlParsed: ...
def validate(url: str, check_domain: bool = ...) -> bool: ...
