.flag, dl.option-list>dt>kbd {
  margin: 0 0.2rem;
  padding: 0 0.2rem;
  border-radius: 0.2rem;
  border: 1px solid var(--color-foreground-border);
  color: var(--color-foreground-primary);
  vertical-align: text-bottom;

  font-size: var(--font-size--small--3);
  display: inline-block;

  box-shadow: 0 0.0625rem 0 rgba(0, 0, 0, 0.2), inset 0 0 0 0.125rem var(--color-background-primary);

  background-color: var(--color-background-secondary);
}

.sig + .sig {
  border-top-left-radius: 0;
  border-top-right-radius: 0;
  padding-top: .05rem;
}
.sig:has(+.sig) {
  border-bottom-left-radius: 0;
  border-bottom-right-radius: 0;
  padding-bottom: .05rem;
}
